@@ -2,7 +2,7 @@
  * Copyright (c) 2004 Atheros Communications, Inc.
  * All rights reserved
  *
- * $Id: if_ath_ahb.c 2833 2007-11-06 00:33:49Z scottr $
+ * $Id: if_ath_ahb.c 3052 2007-12-13 04:09:32Z proski $
  */
 #include "opt_ah.h"
 
@@ -19,7 +19,9 @@
 #include <linux/if.h>
 #include <linux/netdevice.h>
 #include <linux/cache.h>
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,19)
 #include <linux/platform_device.h>
+#endif
 
 #include <asm/io.h>
 #include <asm/uaccess.h>
@@ -32,10 +34,6 @@
 #include "if_ath_ahb.h"
 #include "ah_soc.h"
 
-#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
-#error "Kernel versions older than 2.6.19 are not supported!"
-#endif
-
 struct ath_ahb_softc {
 	struct ath_softc	aps_sc;
 #ifdef CONFIG_PM
@@ -46,6 +44,106 @@ struct ath_ahb_softc {
 static struct ath_ahb_softc *sclist[2] = {NULL, NULL};
 static u_int8_t num_activesc = 0;
 
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
+static struct ar531x_boarddata *ar5312_boardConfig = NULL;
+static char *radioConfig = NULL;
+
+static int
+ar5312_get_board_config(void)
+{
+	int dataFound;
+	char *bd_config;
+	
+	/*
+	 * Find start of Board Configuration data, using heuristics:
+	 * Search back from the (aliased) end of flash by 0x1000 bytes
+	 * at a time until we find the string "5311", which marks the
+	 * start of Board Configuration.  Give up if we've searched
+	 * more than 500KB.
+	 */
+	dataFound = 0;
+	for (bd_config = (char *)0xbffff000;
+	     bd_config > (char *)0xbff80000;
+	     bd_config -= 0x1000) {
+		if ( *(int *)bd_config == AR531X_BD_MAGIC) {
+			dataFound = 1;
+			break;
+		}
+	}
+	
+	if (!dataFound) {
+		printk("Could not find Board Configuration Data\n");
+		bd_config = NULL;
+	}
+	ar5312_boardConfig = (struct ar531x_boarddata *) bd_config;
+	return dataFound;
+}
+
+static int
+ar5312_get_radio_config(void)
+{
+	int dataFound;
+	char *radio_config;
+	
+	/* 
+	 * Now find the start of Radio Configuration data, using heuristics:
+	 * Search forward from Board Configuration data by 0x1000 bytes
+	 * at a time until we find non-0xffffffff.
+	 */
+	dataFound = 0;
+	for (radio_config = ((char *) ar5312_boardConfig) + 0x1000;
+	     radio_config < (char *)0xbffff000;
+	     radio_config += 0x1000) {
+		if (*(int *)radio_config != 0xffffffff) {
+			dataFound = 1;
+			break;
+		}
+	}
+
+	if (!dataFound) { /* AR2316 relocates radio config to new location */
+	    dataFound = 0;
+	    for (radio_config = ((char *) ar5312_boardConfig) + 0xf8;
+			 radio_config < (char *)0xbffff0f8;
+			 radio_config += 0x1000) {
+			if (*(int *)radio_config != 0xffffffff) {
+				dataFound = 1;
+				break;
+			}
+	    }
+	}
+
+	if (!dataFound) {
+		printk("Could not find Radio Configuration data\n");
+		radio_config = NULL;
+	}
+	radioConfig = radio_config;
+	return dataFound;
+}
+
+static int
+ar5312SetupFlash(void)
+{
+	if (ar5312_get_board_config())
+		if (ar5312_get_radio_config())
+			return 1;
+	return 0;
+}	
+
+/*
+ * Read 16 bits of data from offset into *data
+ */
+static void
+ar5312BspEepromRead(u_int32_t off, u_int32_t nbytes, u_int8_t *data)
+{
+	int i;
+	char *eepromAddr = radioConfig;
+	
+	for (i = 0; i < nbytes; i++, off++)
+		data[i] = eepromAddr[off];
+}
+
+#endif
+
 /* set bus cachesize in 4B word units */
 void
 bus_read_cachesize(struct ath_softc *sc, u_int8_t *csz)
@@ -61,16 +159,16 @@ void *
 bus_alloc_consistent(void *hwdev, size_t size, dma_addr_t *dma_handle)
 {
 	void *ret;
-
+     
 	ret = (void *) __get_free_pages(GFP_ATOMIC, get_order(size));
-
+     
 	if (ret != NULL) {
 		memset(ret, 0, size);
 		*dma_handle = __pa(ret);
 		dma_cache_wback_inv((unsigned long) ret, size);
 		ret = UNCAC_ADDR(ret);
 	}
-
+     
 	return ret;
 }
 
@@ -78,7 +176,7 @@ void
 bus_free_consistent(void *hwdev, size_t size, void *vaddr, dma_addr_t dma_handle)
 {
 	unsigned long addr = (unsigned long) vaddr;
-
+    
 	addr = CAC_ADDR(addr);
 	free_pages(addr, get_order(size));
 }
@@ -88,13 +186,13 @@ ahb_enable_wmac(u_int16_t devid, u_int16_t wlanNum)
 {
 	u_int32_t reset;
 	u_int32_t enable;
-
+	
 	if (((devid & AR5315_REV_MAJ_M) == AR5315_REV_MAJ) ||
 		((devid & AR5315_REV_MAJ_M) == AR5317_REV_MAJ)) {
 		u_int32_t reg;
 		u_int32_t *en = (u_int32_t *) AR5315_AHB_ARB_CTL;
-
-		KASSERT(wlanNum == 0, ("invalid wlan # %d", wlanNum));
+		
+		KASSERT(wlanNum == 0, ("invalid wlan # %d", wlanNum)); 
 
 		/* Enable Arbitration for WLAN */
 		*en  |= AR5315_ARB_WLAN;
@@ -108,9 +206,9 @@ ahb_enable_wmac(u_int16_t devid, u_int16_t wlanNum)
 		/* NOTE: for the following write to succeed the
 		 * RST_AHB_ARB_CTL should be set to 0. This driver
 		 * assumes that the register has been set to 0 by boot loader
-		 */
+		 */ 
 		reg = REG_READ(AR5315_PCI_MAC_SCR);
-		reg = (reg & ~AR5315_PCI_MAC_SCR_SLMODE_M) |
+		reg = (reg & ~AR5315_PCI_MAC_SCR_SLMODE_M) | 
 		    (AR5315_PCI_MAC_SCR_SLM_FWAKE << AR5315_PCI_MAC_SCR_SLMODE_S);
 		REG_WRITE(AR5315_PCI_MAC_SCR, reg);
 
@@ -155,11 +253,11 @@ ahb_disable_wmac(u_int16_t devid, u_int16_t wlanNum)
 		((devid & AR5315_REV_MAJ_M) == AR5317_REV_MAJ)) {
 		u_int32_t *en = (u_int32_t *) AR5315_AHB_ARB_CTL;
 
-		KASSERT(wlanNum == 0, ("invalid wlan # %d", wlanNum) );
+		KASSERT(wlanNum == 0, ("invalid wlan # %d", wlanNum) ); 
 
 		/* Enable Arbitration for WLAN */
 		*en &= ~AR5315_ARB_WLAN;
-	} else {
+	} else { 
 		switch (wlanNum) {
 		case AR531X_WLAN0_NUM:
 			enable = AR531X_ENABLE_WLAN0;
@@ -177,22 +275,28 @@ ahb_disable_wmac(u_int16_t devid, u_int16_t wlanNum)
 
 
 static int
-exit_ath_wmac(u_int16_t wlanNum, struct ar531x_config *config)
+exit_ath_wmac(u_int16_t wlanNum)
 {
 	struct ath_ahb_softc *sc = sclist[wlanNum];
 	struct net_device *dev;
-	u_int16_t devid;
+	const char *sysType;
+	u_int16_t devid;        
 
 	if (sc == NULL)
 		return -ENODEV; /* XXX: correct return value? */
-
+        
 	dev = sc->aps_sc.sc_dev;
 	ath_detach(dev);
 	if (dev->irq)
 		free_irq(dev->irq, dev);
-	devid = sc->aps_sc.devid;
-	config->tag = (void *)((unsigned long) devid);
-
+	sysType = get_system_type();
+	if (!strcmp(sysType, "Atheros AR5315"))
+		devid = (u_int16_t) (sysRegRead(AR5315_SREV) &
+			(AR5315_REV_MAJ_M | AR5315_REV_MIN_M));
+	else
+		devid = (u_int16_t) ((sysRegRead(AR531X_REV) >> 8) & 
+			(AR531X_REV_MAJ | AR531X_REV_MIN));
+  
 	ahb_disable_wmac(devid, wlanNum);
 	free_netdev(dev);
 	sclist[wlanNum] = NULL;
@@ -205,11 +309,11 @@ init_ath_wmac(u_int16_t devid, u_int16_t wlanNum, struct ar531x_config *config)
 	const char *athname;
 	struct net_device *dev;
 	struct ath_ahb_softc *sc;
-
+        
 	if (((wlanNum != 0) && (wlanNum != 1)) ||
 		(sclist[wlanNum] != NULL))
 		goto bad;
-
+        
 	ahb_enable_wmac(devid, wlanNum);
 
 	dev = alloc_netdev(sizeof(struct ath_ahb_softc), "wifi%d", ether_setup);
@@ -225,7 +329,11 @@ init_ath_wmac(u_int16_t devid, u_int16_t wlanNum, struct ar531x_config *config)
 	 * interrupts until setup is complete.
 	 */
 	sc->aps_sc.sc_invalid = 1;
+#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,5,41)
+	dev->owner = THIS_MODULE;
+#else
 	SET_MODULE_OWNER(dev);
+#endif
 	sclist[wlanNum] = sc;
 
 	switch (wlanNum) {
@@ -250,11 +358,11 @@ init_ath_wmac(u_int16_t devid, u_int16_t wlanNum, struct ar531x_config *config)
 	sc->aps_sc.sc_iobase = (void __iomem *) dev->mem_start;
 	sc->aps_sc.sc_bdev = NULL;
 
-	if (request_irq(dev->irq, ath_intr, SA_SHIRQ, dev->name, dev)) {
+	if (request_irq(dev->irq, ath_intr, IRQF_SHARED, dev->name, dev)) {
 		printk(KERN_WARNING "%s: request_irq failed\n", dev->name);
 		goto bad3;
 	}
-
+	
 	if (ath_attach(devid, dev, config) != 0)
 		goto bad4;
 	athname = ath_hal_probe(ATHEROS_VENDOR_ID, devid);
@@ -265,7 +373,7 @@ init_ath_wmac(u_int16_t devid, u_int16_t wlanNum, struct ar531x_config *config)
 
 	sc->aps_sc.sc_invalid = 0;
 	return 0;
-
+        
  bad4:
 	free_irq(dev->irq, dev);
  bad3:
@@ -277,22 +385,23 @@ init_ath_wmac(u_int16_t devid, u_int16_t wlanNum, struct ar531x_config *config)
 	return -ENODEV;
 }
 
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,19)
 static int ahb_wmac_probe(struct platform_device *pdev)
 {
-	u_int16_t devid;
+	u32 devid;
 	struct ar531x_config *config;
 
 	config = (struct ar531x_config *) pdev->dev.platform_data;
 	devid = (long) config->tag;
 	config->tag = NULL;
-
-	return init_ath_wmac(devid, pdev->id, config);
+	
+	return init_ath_wmac((u_int16_t) devid, pdev->id, config);
 }
 
 
 static int ahb_wmac_remove(struct platform_device *pdev)
 {
-	exit_ath_wmac(pdev->id, (struct ar531x_config *) pdev->dev.platform_data);
+	exit_ath_wmac(pdev->id);
 
 	return 0;
 }
@@ -303,12 +412,66 @@ static struct platform_driver ahb_wmac_driver = {
 	.remove = ahb_wmac_remove
 };
 
+#else
+
+static int
+init_ahb(void)
+{
+	int ret;
+	u_int16_t devid, radioMask;
+	const char *sysType;
+	struct ar531x_config config;
+
+	sysType = get_system_type();
+	
+	/* Probe to find out the silicon revision and enable the
+	   correct number of macs */
+	if (!ar5312SetupFlash())
+		return -ENODEV;
+
+	config.board = ar5312_boardConfig;
+	config.radio = radioConfig;
+	config.unit = wlanNum;
+	config.tag = NULL;
+
+	if (!strcmp(sysType,"Atheros AR5315")) {
+		devid = (u_int16_t) (sysRegRead(AR5315_SREV) &
+			(AR5315_REV_MAJ_M | AR5315_REV_MIN_M));
+		if (((devid & AR5315_REV_MAJ_M) == AR5315_REV_MAJ) ||
+			((devid & AR5315_REV_MAJ_M) == AR5317_REV_MAJ))
+			return init_ath_wmac(devid, 0, &config);
+	}
+
+	devid = (u_int16_t) ((sysRegRead(AR531X_REV) >>8) &
+		(AR531X_REV_MAJ | AR531X_REV_MIN));
+	switch (devid) {
+	case AR5212_AR5312_REV2:
+	case AR5212_AR5312_REV7:
+		/* Need to determine if we have a 5312 or a 2312 since they
+		 * have the same Silicon Rev ID */
+		ar5312BspEepromRead(2 * AR531X_RADIO_MASK_OFF, 2,
+			(char *) &radioMask);
+		if ((radioMask & AR531X_RADIO0_MASK) != 0)
+			if ((ret = init_ath_wmac(devid, 0, &config)) !=0 )
+				return ret;
+		/* XXX: Fall through?! */
+	case AR5212_AR2313_REV8:
+		if ((ret = init_ath_wmac(devid, 1, &config)) != 0)
+			return ret;
+		break;
+	default:
+		return -ENODEV;
+	}
+	return 0;
+}
+
+#endif
+
 /*
  * Module glue.
  */
-#include "version.h"
 #include "release.h"
-static char *version = ATH_PCI_VERSION " (" RELEASE_VERSION ")";
+static char *version = RELEASE_VERSION;
 static char *dev_info = "ath_ahb";
 
 #include <linux/ethtool.h>
@@ -341,9 +504,17 @@ static int __init
 init_ath_ahb(void)
 {
 	printk(KERN_INFO "%s: %s\n", dev_info, version);
-	platform_driver_register(&ahb_wmac_driver);
-	ath_sysctl_register();
 
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,19)
+	platform_driver_register(&ahb_wmac_driver);
+#else
+	if (init_ahb() != 0) {
+		printk("ath_ahb: No devices found, driver not installed.\n");
+		return (-ENODEV);
+	}
+#endif
+
+	ath_sysctl_register();
 	return 0;
 }
 module_init(init_ath_ahb);
@@ -352,7 +523,14 @@ static void __exit
 exit_ath_ahb(void)
 {
 	ath_sysctl_unregister();
+
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,19)
 	platform_driver_unregister(&ahb_wmac_driver);
+#else
+	exit_ath_wmac(AR531X_WLAN0_NUM);
+	exit_ath_wmac(AR531X_WLAN1_NUM);
+#endif
+
 	printk(KERN_INFO "%s: driver unloaded\n", dev_info);
 }
 module_exit(exit_ath_ahb);
